/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */

function __rest(s, e) {
  var t = {};
  for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0) t[p] = s[p];
  if (s != null && typeof Object.getOwnPropertySymbols === 'function')
    for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
      if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i])) t[p[i]] = s[p[i]];
    }
  return t;
}

function __values(o) {
  var s = typeof Symbol === 'function' && Symbol.iterator,
    m = s && o[s],
    i = 0;
  if (m) return m.call(o);
  if (o && typeof o.length === 'number')
    return {
      next: function() {
        if (o && i >= o.length) o = void 0;
        return { value: o && o[i++], done: !o };
      },
    };
  throw new TypeError(s ? 'Object is not iterable.' : 'Symbol.iterator is not defined.');
}

function __read(o, n) {
  var m = typeof Symbol === 'function' && o[Symbol.iterator];
  if (!m) return o;
  var i = m.call(o),
    r,
    ar = [],
    e;
  try {
    while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
  } catch (error) {
    e = { error: error };
  } finally {
    try {
      if (r && !r.done && (m = i['return'])) m.call(i);
    } finally {
      if (e) throw e.error;
    }
  }
  return ar;
}

/**
 * Utils
 */
var hasOwnProperty = Object.prototype.hasOwnProperty;
function hasOwn(obj, key) {
  return hasOwnProperty.call(obj, key);
}
function warn(condition, msg, vm) {
  if (!condition) {
    // eslint-disable-next-line no-console
    console.warn('[@vue-async] ' + msg, vm);
  }
}
function error(condition, msg, vm) {
  if (!condition) {
    // eslint-disable-next-line no-console
    console.error('[@vue-async] ' + msg, vm);
  }
}
function print(title, content) {
  // eslint-disable-next-line no-console
  console.log(
    '%c' + title + '%c' + content,
    'background: #00d1b2; padding: 5px; color: #fff; border-radius: 5px 0 0 5px',
    'background: #555; padding: 5px; color: #fff; border-radius: 0 5px 5px 0',
  );
}
var toString = function(x) {
  return Object.prototype.toString.call(x);
};
function isPlainObject(x) {
  return toString(x) === '[object Object]';
}

var currentVue = null;
function setCurrentVue(vue) {
  currentVue = vue;
}

function sanitizeComponent(Component) {
  // If Component already sanitized
  if (Component.options && Component._Ctor === Component) {
    return Component;
  }
  if (!Component.options) {
    Component = currentVue.extend(Component);
    Component._Ctor = Component;
  } else {
    Component._Ctor = Component;
    Component.extendOptions = Component.options;
  }
  // For debugging purpose
  if (!Component.options.name && Component.options.__file) {
    Component.options.name = Component.options.__file;
  }
  return Component;
}
function getComponentName(Component) {
  var component = sanitizeComponent(Component);
  return component.options.name;
}

var isProduction = process.env.NODE_ENF === 'production';
// eslint-disable-next-line @typescript-eslint/no-unused-vars
function dynamicComponent(Vue, store) {
  return {
    /**
     * add component to store
     * @param component vue component
     * @param position position
     */
    add: function(component, position) {
      if (!(typeof component === 'object' || typeof component === 'function')) {
        return error(
          isProduction,
          'dynamicComponent function "add" only accept object or function component as param.',
        );
      }
      store.commit('dynamicComponent/add', {
        component: component,
        position: position,
      });
    },
    /**
     * remove component from store
     * @param name component name
     * @param position position
     */
    remove: function(name, position) {
      if (typeof name !== 'string') {
        return error(isProduction, 'dynamicComponent function "remoce" only accept string name of component as param.');
      }
      store.commit('dynamicComponent/remove', {
        name: name,
        position: position,
      });
    },
  };
}

var isProduction$1 = process.env.NODE_ENF === 'production';
var dynamicComponentState = {
  namespaced: true,
  state: {
    GLOBAL: {},
  },
  mutations: {
    // 注册插槽组件
    add: function(state, _a) {
      var component = _a.component,
        _b = _a.position,
        position = _b === void 0 ? 'GLOBAL' : _b;
      var _component = component.component || component;
      // 必须使用外部 Vue 对象，本地 import 对象不一致
      var componentName = component.name || getComponentName(_component);
      if (!componentName) {
        return error(isProduction$1, 'component must be named for dynamic component.');
      }
      if (!state[position]) {
        state[position] = {};
      }
      if (state[position][componentName]) {
        return error(
          isProduction$1,
          'async component "' + componentName + '" has existed in the ' + position + ', ignored it.',
        );
      }
      state[position][componentName] = component;
    },
    // 移除注册插槽组件
    remove: function(state, _a) {
      var name = _a.name,
        _b = _a.position,
        position = _b === void 0 ? 'GLOBAL' : _b;
      if (!state[position]) {
        return error(isProduction$1, 'position "' + position + '" does not  exist.');
      }
      if (!state[position][name]) {
        return error(isProduction$1, 'async component "' + name + '" does not  exist in the ' + position + '.');
      }
      delete state[position][name];
    },
  },
};

var isProduction$2 = process.env.NODE_ENF === 'production';
// eslint-disable-next-line @typescript-eslint/no-unused-vars
function eventBus(Vue) {
  var events = {};
  return {
    /**
     * 发射事件
     * @param {string} eventName 事件名称
     * @param {any} payload 事件载荷
     */
    emit: function(eventName, payload) {
      var e_1, _a;
      if (!events[eventName]) {
        return warn(
          isProduction$2,
          'event "' + eventName + '" is triggered, but no listeners\uFF0Cnothing will be happening.',
        );
      }
      try {
        for (var _b = __values(events[eventName].values()), _c = _b.next(); !_c.done; _c = _b.next()) {
          var func = _c.value;
          func(payload);
        }
      } catch (e_1_1) {
        e_1 = { error: e_1_1 };
      } finally {
        try {
          if (_c && !_c.done && (_a = _b.return)) _a.call(_b);
        } finally {
          if (e_1) throw e_1.error;
        }
      }
    },
    /**
     * 事件监听
     * @param eventName 事件名称
     * @param handler 处理事件的方法
     */
    on: function(eventName, handler) {
      if (typeof handler === 'function') {
        if (!events[eventName]) events[eventName] = new Set();
        events[eventName].add(handler);
      } else {
        error(isProduction$2, 'handler must be a function, but recived ' + typeof handler);
      }
    },
    /**
     * 取消事件监听
     * @param eventName 事件名称
     * @param handler 处理事件的方法
     */
    off: function(eventName, handler) {
      if (!handler) {
        return warn(!isProduction$2, 'handler must be a valid event function');
      } else if (!events.hasOwnProperty(eventName)) {
        return error(isProduction$2, 'no event name of "' + eventName + '" is on listening');
      }
      events[eventName].delete(handler);
    },
    /**
     * 清理事件总线
     */
    clear: function() {
      events = {};
    },
    /**
     * 获取当前事件总线详情
     */
    getEvents: function() {
      return events;
    },
  };
}

var isProduction$3 = 'development' === 'production';
var moduleLoader = function(Vue, status) {
  return function loader(modules, _a) {
    var _this = this;
    var _b = (_a === void 0 ? {} : _a).__isArray__,
      __isArray__ = _b === void 0 ? false : _b;
    status.current = false;
    if (Array.isArray(modules) && modules.length) {
      return Promise.all(
        modules.map(function(module) {
          return loader.call(_this, module, { __isArray__: true });
        }),
      ).then(function() {
        status.current = true;
      });
    } else if (typeof modules === 'object' && Object.getPrototypeOf(modules) === Object.prototype) {
      /** 通过模块清单加载模块 */
      var promiseAll = [];
      var _loop_1 = function(moduleName) {
        if (!window[moduleName]) {
          promiseAll.push(
            new Promise(function(resolve) {
              var script = document.createElement('script');
              script.src = modules[moduleName];
              script.onload = function() {
                if (window[moduleName]) {
                  typeof window[moduleName] === 'function'
                    ? window[moduleName].call(_this, Vue)
                    : window[moduleName].default.call(_this, Vue);
                  {
                    print(moduleName, 'loaded');
                  }
                } else {
                  warn(isProduction$3, 'module "' + moduleName + '" loaded failed\uFF0Cignored it.');
                }
                resolve();
              };
              script.onerror = function() {
                warn(isProduction$3, 'module "' + moduleName + '" had a problem to create, ignored it.');
                resolve();
              };
              document.body.appendChild(script);
            }),
          );
        }
      };
      for (var moduleName in modules) {
        _loop_1(moduleName);
      }
      return Promise.all(promiseAll).then(function() {
        if (!__isArray__) {
          status.current = true;
        }
      });
    } else if (typeof modules === 'function') {
      /** 通过模块函数加载模块 */
      var result = modules.call(this, Vue);
      if (result && result instanceof Promise) {
        return result.then(function() {
          if (!__isArray__) {
            status.current = true;
          }
        });
      } else {
        if (!__isArray__) {
          status.current = true;
        }
        return Promise.resolve();
      }
    } else {
      error(isProduction$3, '模块加载方法只接受模块列表对象或者模块函数对象作为参数。');
      if (!__isArray__) {
        status.current = true;
      }
      return Promise.resolve();
    }
  };
};

function install(Vue, options) {
  if (options === void 0) {
    options = {};
  }
  if (install.installed) return;
  install.installed = true;
  // 用于监听模块加载的情况，处理路由404问题
  var vm = new Vue({
    data: function() {
      return {
        status: {
          current: false,
        },
      };
    },
  });
  Object.defineProperties(Vue.prototype, {
    $eventBus: {
      value: eventBus(),
      writable: false,
    },
    $moduleLoader: {
      value: moduleLoader(Vue, vm.status),
      writable: false,
    },
  });
  var store = options.store,
    router = options.router;
  // store
  if (store) {
    store.registerModule('dynamicComponent', dynamicComponentState);
    // define $dynamicComponent
    Object.defineProperty(Vue.prototype, '$dynamicComponent', {
      value: dynamicComponent(Vue, store),
      writable: false,
    });
  }
  // router
  // 解决动态路由404问题
  if (router) {
    var resolveRoute_1 = function(to, next) {
      var fullPath = to.redirectedFrom || to.fullPath;
      var _a = router.resolve(fullPath),
        resolved = _a.resolved,
        location = _a.location;
      // 在加载完组件后resolve的地址与原来要跳转的地址不一致时跳转
      // 以免造成404死循环
      if (resolved.name !== to.name) {
        next(location);
      } else {
        next();
      }
    };
    router.beforeEach(function(to, from, next) {
      if (!to.name || to.name === '404' || to.name === 'page-not-found') {
        // 模块已经被加载完成, 但由于在其之前添加 beforeEach 阻止时间过长，vm.$watch还没开始监听
        if (vm.status.current) {
          resolveRoute_1(to, next);
        }
        vm.$watch(
          function() {
            return vm.status.current;
          },
          function(newVal, oldVal) {
            // false => true
            if (newVal && !oldVal) {
              resolveRoute_1(to, next);
            }
          },
        );
      } else {
        next();
      }
    });
  }
  // Used to avoid multiple mixins being setup
  // when in dev mode and hot module reload
  // https://github.com/vuejs/vue/issues/5089#issuecomment-284260111
  if (Vue.$__module_loader_installed__) return;
  // eslint-disable-next-line @typescript-eslint/camelcase
  Vue.$__module_loader_installed__ = true;
  // 设置当前运行时 Vue 对象
  setCurrentVue(Vue);
  // 设置子模块中的运行时 Vue 对象与主框架一致
  // 或使用 '--inline-vue' 使用独立 Vue 对象
  if (!window.Vue) {
    window.Vue = Vue;
  }
  Vue.mixin({
    beforeCreate: function() {
      var options = this.$options;
      if (options.moduleLoader) {
        options.moduleLoader.init(this, options.ssrContext);
        this.$moduleLoadManager = Vue.observable(options.moduleLoader.framework);
        // 初始化实例中的 modules
        // 放在最后执行，module 入口可以使用 this.$moduleLoadManager中的功能
        options.moduleLoader.initModules(this).finally(function() {
          // 在没有模块被加载时，这个值必须要手动改成 true
          // 使 router.beforeEach 正确执行
          if (!vm.status.current) {
            vm.status.current = true;
          }
        });
      } else {
        this.$moduleLoadManager = (options.parent && options.parent.$moduleLoadManager) || this;
      }
    },
  });
}

var ModuleLoader = /** @class */ (function() {
  function ModuleLoader(_a) {
    if (_a === void 0) {
      _a = {};
    }
    var modules = _a.modules,
      rest = __rest(_a, ['modules']);
    this.framework = {
      loaded: false,
      layouts: {},
    };
    this.__initModules__ = modules;
    this.__ExtraParams__ = rest;
  }
  // eslint-disable-next-line @typescript-eslint/no-unused-vars
  ModuleLoader.prototype.init = function(root, ssrContext) {
    var _this = this;
    this.__ExtraParams__ &&
      Object.entries(this.__ExtraParams__).map(function(_a) {
        var _b = __read(_a, 2),
          key = _b[0],
          value = _b[1];
        if (!hasOwn(_this.framework, key)) {
          _this.framework[key] = value;
        } else {
          warn('development' === 'production', '\u53C2\u6570 ' + key + ' \u5DF2\u5728\u5728');
        }
      });
    // 默认方法
    !hasOwn(this.framework, 'addRoutes') && (this.framework.addRoutes = this.initAddRoutes(root));
    !hasOwn(this.framework, 'addLayouts') && (this.framework.addLayouts = this.initAddLayouts(root));
  };
  ModuleLoader.prototype.initModules = function(root) {
    var _this = this;
    if (this.__initModules__) {
      return root.$moduleLoader(this.__initModules__).then(function() {
        _this.framework.loaded = true;
      });
    } else {
      this.framework.loaded = true;
      return Promise.resolve();
    }
  };
  ModuleLoader.prototype.initAddRoutes = function(root) {
    return function(routes) {
      if (root.$router) {
        root.$router.addRoutes(routes);
      } else {
        error('development' === 'production', 'vm.$router未定义');
      }
    };
  };
  ModuleLoader.prototype.initAddLayouts = function(root) {
    var _this = this;
    return function(key, layout) {
      if (typeof key === 'string') {
        root.$set(_this.framework.layouts, key, layout);
      } else if (isPlainObject(key)) {
        // plain object
        var layouts = key;
        _this.framework.layouts = Object.assign(_this.framework.layouts, layouts);
      } else {
        error('development' === 'production', 'addLayouts中支持字符串(key)或 Object(key/value)格式');
      }
    };
  };
  ModuleLoader.install = install;
  // static installed = false
  ModuleLoader.version = '0.1.3-alpha.1';
  return ModuleLoader;
})();

var install$1 = ModuleLoader.install;
ModuleLoader.install = function(Vue, options) {
  if (options === void 0) {
    options = {};
  }
  install$1.call(ModuleLoader, Vue, options);
};
if (typeof window !== 'undefined' && window.Vue) {
  window.Vue.use(ModuleLoader);
}

export default ModuleLoader;
